import json
import logging

import basyx.aas.model
from spade.behaviour import OneShotBehaviour

from smia import GeneralUtils
from smia.logic import inter_aas_interactions_utils
from smia.logic.exceptions import CapabilityRequestExecutionError, CapabilityCheckingError, RequestDataError, \
    AssetConnectionError, OntologyReadingError, AASModelReadingError
from smia.css_ontology.css_ontology_utils import CapabilitySkillACLInfo
from smia.utilities import smia_archive_utils
from smia.utilities.smia_info import AssetInterfacesInfo
from smia.utilities.fipa_acl_info import FIPAACLInfo, ACLJSONSchemas

_logger = logging.getLogger(__name__)


class HandleCapabilityBehaviour(OneShotBehaviour):
    """
    This class implements the behaviour that handles a request related to the capabilities of the Digital Twin.
    """

    def __init__(self, agent_object, svc_req_data):
        """
        The constructor method is rewritten to add the object of the agent.

        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
            svc_req_data (dict): all the information about the service request
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object
        self.svc_req_data = svc_req_data

        self.requested_timestamp = GeneralUtils.get_current_timestamp()

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("HandleCapabilityBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # First, the performative of request is obtained and, depending on it, different actions will be taken
        match self.svc_req_data['performative']:
            case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:  # TODO actualizar dentro de todo el codigo los usos de performativas y ontologias de FIPA-ACL
                await self.handle_request()
            case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF:
                await self.handle_query_if()
            case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM:
                await self.handle_inform()
            case "PensarOtro":  # TODO
                pass
            case _:
                _logger.error("Performative not available for capability management.")
        self.exit_code = 0

    # ------------------------------------------
    # Methods to handle of all types of services
    # ------------------------------------------
    async def handle_request(self):
        """
        This method handle capability requests to the DT.
        """

        # TODO, para este paso, se podrian almacenar las capacidades que se han verificado ya cuando se recibe el
        #  Query-If (se supone que otro agente debería mandar en CallForProposal despues del Query-If, pero para
        #  añadirle una validacion extra) Esto podria hacerse con el thread (el Query-If y CFP estarían en la misma
        #  negociacion?). Otra opcion es siempre ejecutar aas_model.capability_checking_from_acl_request()
        cap_name = None
        try:
            # First, the data received is checked to ensure that it contains all the necessary information.
            await self.check_received_capability_request_data()

            # The instances of capability, skill and skill interface are obtained depending on the received data
            cap_ontology_instance, skill_ontology_instance, skill_interface_ontology_instance = \
                await self.get_ontology_instances()
            cap_name = cap_ontology_instance.name

            # Before executing the capability, the capability checking should be also performed to ensure that SMIA
            # can perform it.
            result, reason = await self.execute_capability_checking()
            if not result:
                raise CapabilityRequestExecutionError(
                    cap_name,f"The capability {cap_name} cannot be executed because the capability checking "
                             f"result is invalid. Reason: {reason}.", self)

            # Once all the data has been checked and obtained, the capability can be executed
            cap_execution_result = await self.execute_capability(cap_ontology_instance, skill_ontology_instance,
                                                                 skill_interface_ontology_instance)

            # When the skill has finished, the request is answered
            await self.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
                                                   {'result': cap_execution_result})

            _logger.info("Management of the capability {} finished.".format(cap_name))

            # The information will be stored in the log
            execution_info = {'capName': cap_name, 'capType': str(cap_ontology_instance.is_a),
                              'result': str(cap_execution_result), 'taskType': 'CapabilityRequest'}
            smia_archive_utils.save_completed_svc_log_info(self.requested_timestamp,
                                                           GeneralUtils.get_current_timestamp(),
                                                           self.svc_req_data, execution_info,
                                                           self.svc_req_data['serviceType'])

        except (RequestDataError, OntologyReadingError, CapabilityRequestExecutionError,
                AASModelReadingError, AssetConnectionError) as cap_request_error:
            if isinstance(cap_request_error, RequestDataError):
                if 'JSON schema' in cap_request_error.message:
                    cap_name = ''
            if isinstance(cap_request_error, RequestDataError) or isinstance(cap_request_error, OntologyReadingError) \
                    or isinstance(cap_request_error, AASModelReadingError):
                cap_request_error = CapabilityRequestExecutionError(cap_name, cap_request_error.__class__.__name__ +
                                                                    ': ' + cap_request_error.message, self)
            if isinstance(cap_request_error, AssetConnectionError):
                cap_request_error = CapabilityRequestExecutionError(
                    cap_name, f"The error [{cap_request_error.error_type}] has appeared during the asset "
                              f"connection. Reason: {cap_request_error.reason}.", self)

            await cap_request_error.handle_capability_execution_error()
            return  # killing a behaviour does not cancel its current run loop

    async def handle_query_if(self):
        """
        This method handle Query-If requests for the Capability. This request is received when the DT is asked about
        information related to a capability.

        """
        # TODO modificar el codigo para usar la ontologia (sigue con el enfoque antiguo)
        cap_name = None
        try:
            # First, the data received is checked to ensure that it contains all the necessary information.
            await self.check_received_capability_request_data()     # TODO de momento se analiza igual que las Capability requests (mas adelante pensar como son los mensajes ACL para CapabilityRequest y CapabilityChecking)

            # Then, the capability checking process can be executed
            result, reason = await self.execute_capability_checking()

            # When the checking has finished, the request is answered
            cap_name = self.svc_req_data['serviceData']['serviceParams'][CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME]
            cap_ontology_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(cap_name)
            await self.send_response_msg_to_sender(FIPAACLInfo.FIPA_ACL_PERFORMATIVE_INFORM,
                                                   {'result': result, 'reason': reason})

            _logger.info("Checking of the capability {} finished with result {}.".format(cap_name, result))

            # The information will be stored in the log
            execution_info = {'capName': cap_name, 'capType': str(cap_ontology_instance.is_a),
                              'result': str(result), 'reason': reason, 'taskType': 'CapabilityChecking'}
            smia_archive_utils.save_completed_svc_log_info(self.requested_timestamp,
                                                           GeneralUtils.get_current_timestamp(),
                                                           self.svc_req_data, execution_info,
                                                           self.svc_req_data['serviceType'])

        except (RequestDataError, AASModelReadingError) as cap_checking_error:
            if isinstance(cap_checking_error, RequestDataError):
                if 'JSON schema' in cap_checking_error.message:
                    cap_name = ''
            cap_checking_error = CapabilityCheckingError(cap_name, cap_checking_error.__class__.__name__ +
                                                                ': ' + cap_checking_error.message, self)
            await cap_checking_error.handle_capability_checking_error()
            return  # killing a behaviour does not cancel its current run loop

    async def handle_inform(self):
        """
        This method handles AAS Services. These services serve for the management of asset-related information through
        a set of infrastructure services provided by the AAS itself. These include Submodel Registry Services (to list
        and register submodels), Meta-information Management Services (including Classification Services, to check if the
        interface complies with the specifications; Contextualization Services, to check if they belong together in a
        context to build a common function; and Restriction of Use Services, divided between access control and usage
        control) and Exposure and Discovery Services (to search for submodels or asset related services).

        """
        _logger.info('CSS-related INFORM performative message: ' + str(self.svc_req_data))


    # Internal logic methods
    # ----------------------
    async def send_response_msg_to_sender(self, performative, service_params):
        """
        This method creates and sends a FIPA-ACL message with the given serviceParams and performative.

        Args:
            performative (str): performative according to FIPA-ACL standard.
            service_params (dict): JSON with the serviceParams to be sent in the message.
        """
        acl_msg = inter_aas_interactions_utils.create_inter_smia_response_msg(
            receiver=self.svc_req_data['sender'],
            thread=self.svc_req_data['thread'],
            performative=performative,
            ontology='CapabilityResponse',
            service_id=self.svc_req_data['serviceID'],
            service_type=self.svc_req_data['serviceType'],
            service_params=json.dumps(service_params)
        )
        await self.send(acl_msg)

    async def check_received_capability_request_data(self):
        """
        This method checks whether the data received contains the necessary information to be able to execute
        the capability. If an error occurs, it throws a CapabilityDataError exception.
        """
        # TODO con el nuevo enfoque solo es necesario añadir la capacidad. Las demas son opcionales (skill, interfaz...)
        #  Esto se debe a que una capacidad puede solicitarse sin definir skills (si ninguna tiene parametros, SMIA
        #  ejecutaría la capacidad con una skill aleatoria). Lo que si se va a comprobar es que si se añaden datos
        #  opcionales, sean validos (p.e. que los datos añadidos estén conformes a la ontologia CSS)
        # First, the structure and attributes of the received data are checked and validated
        await inter_aas_interactions_utils.check_received_request_data_structure(
            self.svc_req_data, ACLJSONSchemas.JSON_SCHEMA_CAPABILITY_REQUEST)
        received_cap_data = self.svc_req_data['serviceData']['serviceParams']
        # if CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME not in received_cap_data:
        #     raise RequestDataError("The received capability data is invalid due to missing #{} field in request "
        #                               "message.".format(CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME))
        # else:
        cap_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME]
        cap_ontology_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(cap_name)
        if cap_ontology_instance is None:
            raise RequestDataError("The capability {} does not an instance defined in the ontology of this "
                                   "DT".format(cap_name))
        if CapabilitySkillACLInfo.REQUIRED_SKILL_NAME in received_cap_data:
            skill_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_NAME]
            result, skill_instance = cap_ontology_instance.check_and_get_related_instance_by_instance_name(skill_name)
            if result is False:
                raise RequestDataError("The capability {} and skill {} are not linked in the ontology of this "
                                       "DT, or the skill does not have an instance"
                                       ".".format(cap_name, skill_name))
            if skill_instance.get_associated_skill_parameter_instances() is not None:
                # It is checked that the necessary parameter data have been added (in this case only the input
                # parameters will be necessary).
                skill_params = skill_instance.get_associated_skill_parameter_instances()
                for param in skill_params:
                    if param.is_skill_parameter_type(['INPUT', 'INOUTPUT']):
                        # If there is a parameter of type INPUT or INOUTPUT the value of the parameter need to be
                        # specified in the request message
                        if CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS_VALUES not in received_cap_data:
                            raise RequestDataError("The received request is invalid due to missing #{} field in the "
                                                   "request message because the requested skill need value for an input"
                                                   " parameter ({}).".format(
                                CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS_VALUES, param.name))
                        if param.name not in received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS_VALUES].keys():
                            raise RequestDataError("The received request is invalid due to missing #{} field in the "
                                                   "request message because the requested skill need value for an input"
                                                   " parameter ({}).".format(
                                CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS_VALUES, param.name))
            if CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME in received_cap_data:
                # Solo si se ha definido la skill se define la skill interface, sino no tiene significado
                # TODO pensar la frase anterior. Realmente tiene significado o no? Si no se define la skill, podriamos
                #  definir una interfaz que queremos utilizar si o si? En ese caso, habria que buscar una skill con esa
                #  interfaz para ejecutarla
                skill_interface_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME]
                result, instance = skill_instance.check_and_get_related_instance_by_instance_name(skill_interface_name)
                if result is False:
                    raise RequestDataError("The skill {} and skill interface {} are not linked in the ontology of "
                                           "this DT, or the skill interface does not have an instance"
                                           ".".format(skill_name, skill_interface_name))
        # The constraints of the given capability are also checked
        constraint_instances = cap_ontology_instance.get_associated_constraint_instances()
        if constraint_instances is not None:
            if CapabilitySkillACLInfo.REQUIRED_CAPABILITY_CONSTRAINTS not in received_cap_data:
                raise RequestDataError("The received request is invalid because the #{} field is missing in the "
                                       "request message, and the given capability {} has constraints defined.".format(
                    CapabilitySkillACLInfo.REQUIRED_CAPABILITY_CONSTRAINTS, cap_name))
            for constraint in constraint_instances:
                if constraint.name not in received_cap_data[CapabilitySkillACLInfo.REQUIRED_CAPABILITY_CONSTRAINTS].keys():
                    raise RequestDataError("The received request is invalid due to missing #{} field in the"
                                           "request message because the requested capability need value for this "
                                           "constraint ({}).".format(
                        CapabilitySkillACLInfo.REQUIRED_CAPABILITY_CONSTRAINTS, constraint.name))
    async def check_received_skill_data(self, skill_elem):
        """
        This method checks whether the data received contains the necessary information in relation to the skill of the
        received capability request.

        Args:
            skill_elem (basyx.aas.model.SubmodelElement): skill Python object in form of a SubmodelElement.

        Returns:
            bool: the result of the check.
        """
        received_skill_data = self.svc_req_data['serviceData']['serviceParams'][
            CapabilitySkillACLInfo.REQUIRED_SKILL_INFO]
        if CapabilitySkillACLInfo.REQUIRED_SKILL_NAME not in received_skill_data:
            raise RequestDataError(
                "The received capability data is invalid due to missing #{} field in the skill "
                "information section of the request message.".format(CapabilitySkillACLInfo.REQUIRED_SKILL_NAME))
        if CapabilitySkillACLInfo.REQUIRED_SKILL_ELEMENT_TYPE not in received_skill_data:
            raise RequestDataError(
                "The received capability data is invalid due to missing #{} field in the skill "
                "information section of the request message.".format(
                    CapabilitySkillACLInfo.REQUIRED_SKILL_ELEMENT_TYPE))
        # If the skill has parameters, it will be checked if they exist within the received data
        # TODO de momento los skills son solo Operation, pensar como recoger los skill parameters para los demas casos
        if skill_elem.input_variable or skill_elem.output_variable:
            if CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS not in received_skill_data:
                raise RequestDataError(
                    "The received capability data is invalid due to missing #{} field in the skill information"
                    " section of the request message.".format(CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS))
        if skill_elem.input_variable:
            if CapabilitySkillACLInfo.REQUIRED_SKILL_INPUT_PARAMETERS not in received_skill_data[
                CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS]:
                raise RequestDataError(
                    "The received capability data is invalid due to missing #{} field in the skill parameters "
                    "information section of the request message.".format(
                        CapabilitySkillACLInfo.REQUIRED_SKILL_INPUT_PARAMETERS))
        if skill_elem.output_variable:
            if CapabilitySkillACLInfo.REQUIRED_SKILL_OUTPUT_PARAMETERS not in received_skill_data[
                CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS]:
                raise RequestDataError(
                    "The received capability data is invalid due to missing #{} field in the skill parameters "
                    "information section of the request message.".format(
                        CapabilitySkillACLInfo.REQUIRED_SKILL_OUTPUT_PARAMETERS))

    async def get_ontology_instances(self):
        """
        This method gets the ontology instances for the capability, skill and skill interface depending on the received
        data. If the data is invalid or there are no available combination of three instances, it raises an Exception.

        Returns:
            capability_instance (owlready2.ThingClass), skill_instance (owlready2.ThingClass),
            skill_interface_instance (owlready2.ThingClass): ontology instances for capability, skill and skill
            interface.
        """

        received_cap_data = self.svc_req_data['serviceData']['serviceParams']
        cap_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME]
        cap_ontology_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(cap_name)

        # The CSS ontology is used to search for instances
        if CapabilitySkillACLInfo.REQUIRED_SKILL_NAME not in received_cap_data:
            cap_associated_skills = cap_ontology_instance.get_associated_skill_instances()
            if cap_associated_skills is None:
                raise CapabilityRequestExecutionError(cap_name, "The capability {} does not have any associated skill, "
                                                                "so it cannot be executed".format(cap_name), self)
            else:
                # En este caso no se ha determinado una skill en concreto, asi que se comprobará si alguna
                # skill no tiene parametros, por lo que puede ser directamente ejecutada
                for skill in cap_associated_skills:
                    if skill.get_associated_skill_parameter_instances() is None:
                        # En este caso no tiene parametros, por lo que se puede ejecutar directamente
                        if len(list(skill.get_associated_skill_interface_instances())) == 0:
                            # If the skill does not have interfaces, it cannot be executed
                            continue
                        # The first skill interface is obtained to execute the capability
                        return cap_ontology_instance, skill, list(skill.get_associated_skill_interface_instances())[0]
                else:
                    # En este caso todas las skills tienen parametros, asi que hay que solicitarselos al
                    # requester de la ejecución de la capacidad. Se devolverán todas skills con sus
                    # parametros para que vuelva a solicitar la ejecución de la capacidad pero con una skill
                    # especificada y con los valores necesarios de sus parametros
                    raise RequestDataError("To execute the capability {}, the skill and its parameters need to be added"
                                           " in the request message.".format(cap_name))
        else:
            # En este caso se ha especificado una skill, asi que primera se analizara si tiene parametros y
            # si estos se han definido
            result, skill_ontology_instance = (cap_ontology_instance.check_and_get_related_instance_by_instance_name(
                received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_NAME]))
            skill_interface_ontology_instance = None
            if CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME not in received_cap_data:
                # Si no se ha definido una skill interface, se recoge la primera
                skill_interface_ontology_instance = list(skill_ontology_instance.
                                                         get_associated_skill_interface_instances())[0]
                if skill_interface_ontology_instance is None:
                    raise CapabilityRequestExecutionError(cap_name, "The capability requested by the given"
                                                                    " skill cannot be executed because there is no skill "
                                                                    "interface defined.", self)
            else:
                result, skill_interface_ontology_instance = skill_ontology_instance.check_and_get_related_instance_by_instance_name(
                    received_cap_data[CapabilitySkillACLInfo.REQUIRED_SKILL_INTERFACE_NAME])
            return cap_ontology_instance, skill_ontology_instance, skill_interface_ontology_instance

    async def execute_capability_checking(self):
        """
        This method executes the process of checking the requested capability to ensure that this SMIA can perform it.

        Returns:
            bool: result of the capability checking
        """
        # First, the ontology instance of the capability is obtained
        received_cap_data = self.svc_req_data['serviceData']['serviceParams']
        cap_name = received_cap_data[CapabilitySkillACLInfo.REQUIRED_CAPABILITY_NAME]
        cap_ontology_instance = await self.myagent.css_ontology.get_ontology_instance_by_name(cap_name)

        # The associated constraints and received values for them are also obtained
        constraint_instance_list = cap_ontology_instance.get_associated_constraint_instances()
        received_constraint_data = self.svc_req_data['serviceData']['serviceParams'][
            CapabilitySkillACLInfo.REQUIRED_CAPABILITY_CONSTRAINTS]

        # At this point, the data received has already been verified, so that the values of the constraints can be
        # checked directly.
        for constraint_instance in constraint_instance_list:
            aas_cap_constraint_elem = await self.myagent.aas_model.get_object_by_reference(
                constraint_instance.get_aas_sme_ref())
            result = aas_cap_constraint_elem.check_constraint(received_constraint_data[constraint_instance.name])
            if not result:
                return False, 'The constraint {} with data {} is not valid'.format(constraint_instance.name,
                                                                                   received_constraint_data[constraint_instance.name])
        # If all capability constraint are valid, the checking valid
        # TODO PENSAR MAS VALIDACIONES DURANTE EL CAPABILITY CHECKING
        return True, ''

    async def execute_capability(self, cap_instance, skill_instance, skill_interface_instance):
        """
        This method executes a given capability through as an implementation of a given skill through a given skill
        interface. All the data received are instances of the CSS ontology.

        Args:
            cap_instance (owlready2.ThingClass): ontology instance of the capability to execute.
            skill_instance (owlready2.ThingClass): ontology instance of the skill to execute.
            skill_interface_instance (owlready2.ThingClass): ontology instance of the skill interface to use.

        Returns:
            object: result of the capability execution
        """
        aas_cap_elem = await self.myagent.aas_model.get_object_by_reference(cap_instance.get_aas_sme_ref())
        aas_skill_elem = await self.myagent.aas_model.get_object_by_reference(skill_instance.get_aas_sme_ref())
        aas_skill_interface_elem = await self.myagent.aas_model.get_object_by_reference(
            skill_interface_instance.get_aas_sme_ref())
        if None in (aas_cap_elem, aas_skill_elem, aas_skill_interface_elem):
            raise CapabilityRequestExecutionError(cap_instance.name, "The requested capability {} cannot be executed"
                                                                     " because there is no AAS element linked to the ontology "
                                                                     "instances.".format(cap_instance.name), self)

        parent_submodel = aas_skill_interface_elem.get_parent_submodel()
        if parent_submodel.check_semantic_id_exist(AssetInterfacesInfo.SEMANTICID_INTERFACES_SUBMODEL):
            # In this case, the capability need to be executed through an asset service
            # The asset interface will be obtained from the skill interface SubmodelElement.
            aas_asset_interface_elem = aas_skill_interface_elem.get_associated_asset_interface()
            # TODO PENSAR COMO SERIA CON UN AGENT SERVICE
            # With the AAS SubmodelElement of the asset interface the related Python class, able to connect to the asset,
            # can be obtained.
            asset_connection_class = await self.myagent.get_asset_connection_class_by_ref(aas_asset_interface_elem)
            _logger.assetinfo("The Asset connection of the Skill Interface has been obtained.")
            # Now the capability can be executed through the Asset Connection class related to the given skill. The required
            # input data can be obtained from the received message, since it has already been verified as containing such data
            # TODO FALTA DESARROLLARLO EN EL METODO DE CHEQUEO DEL MENSAJE
            _logger.assetinfo("Executing skill of the capability through an asset service...")
            received_skill_input_data = self.svc_req_data['serviceData']['serviceParams'][
                CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS_VALUES]
            skill_execution_result = await asset_connection_class.execute_asset_service(
                interaction_metadata=aas_skill_interface_elem, service_input_data=received_skill_input_data)
            _logger.assetinfo("Skill of the capability successfully executed.")

            # TODO SI LA SKILL TIENE OUTPUT PARAMETERS, HAY QUE RECOGERLOS DEL skill_execution_result. En ese caso, se
            #  sobreescribirá el skill_execution_result con la variable output y su valor (el cual será lo que devolverá el
            #  metodo del asset connnection class)
        else:
            # In this case, the capability need to be executed through an agent service
            # TODO FALTA COMPROBAR QUE FUNCIONA
            try:
                received_skill_input_data = self.svc_req_data['serviceData']['serviceParams'][
                    CapabilitySkillACLInfo.REQUIRED_SKILL_PARAMETERS_VALUES]
                skill_execution_result = await self.myagent.agent_services.execute_agent_service_by_id(
                    aas_skill_interface_elem.id_short, **received_skill_input_data)
            except (KeyError, ValueError) as e:
                raise CapabilityRequestExecutionError(cap_instance.name, "The requested capability {} cannot be "
                                                      "executed because the agent service {} cannot be successfully "
                                                      "executed.".format(cap_instance.name,
                                                      aas_skill_interface_elem.id_short), self)

        return skill_execution_result




class FIPAACLInfo:
    """
    This class contains the values related to FIPA-ACL standard.
    """

    # Performative values
    FIPA_ACL_PERFORMATIVE_CFP = 'CallForProposal'
    FIPA_ACL_PERFORMATIVE_INFORM = 'Inform'
    FIPA_ACL_PERFORMATIVE_REQUEST = 'Request'
    FIPA_ACL_PERFORMATIVE_PROPOSE = 'Propose'
    FIPA_ACL_PERFORMATIVE_FAILURE = 'Failure'
    FIPA_ACL_PERFORMATIVE_QUERY_IF = 'Query-If'
    # TODO add more if they are needed
    # TODO se han añadido estos pero todavia no se utilizan:
    FIPA_ACL_PERFORMATIVE_ACCEPT_PROPOSAL = 'AcceptProposal'
    FIPA_ACL_PERFORMATIVE_REJECT_PROPOSAL = 'RejectProposal'
    FIPA_ACL_PERFORMATIVE_AGREE = 'Agree'
    FIPA_ACL_PERFORMATIVE_CONFIRM = 'Confirm'
    FIPA_ACL_PERFORMATIVE_NOT_UNDERSTOOD = 'NotUnderstood'
    FIPA_ACL_PERFORMATIVE_REFUSE = 'Refuse'

    # Ontology values
    FIPA_ACL_ONTOLOGY_SVC_REQUEST = 'SvcRequest'
    FIPA_ACL_ONTOLOGY_SVC_RESPONSE = 'SvcResponse'
    FIPA_ACL_ONTOLOGY_CAPABILITY_REQUEST = 'CapabilityRequest'
    FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING = 'CapabilityChecking'
    FIPA_ACL_ONTOLOGY_CAPABILITY_RESPONSE = 'CapabilityResponse'
    FIPA_ACL_ONTOLOGY_SVC_NEGOTIATION = 'Negotiation'

class ServiceTypes:
    """
    This class contains all service types defined in the Functional View of RAMI 4.0.
    """
    ASSET_RELATED_SERVICE = 'AssetRelatedService'
    AAS_INFRASTRUCTURE_SERVICE = 'AASInfrastructureService'
    AAS_SERVICE = 'AASservice'
    SUBMODEL_SERVICE = 'SubmodelService'
    CSS_RELATED_SERVICE = 'CSSRelatedService'   # TODO duda con este ya que contiene el concepto de service dentro, y no es lo mismo

class ACLJSONSchemas:
    """This class contains all the JSON schemas related to ACL messages sent between SMIA agents."""

    JSON_SCHEMA_AAS_MODEL_REFERENCE = {
        "type": "object",
        "properties": {
            "keys": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "type": {"type": "string"},
                        "value": {"type": "string"}
                    },
                    "required": ["type", "value"]
                }
            }
        },
        "required": ["keys"]
    }

    JSON_SCHEMA_SUBMODEL_SERVICE_REQUEST = {
        "type": "object",
        "properties": {
            "ModelReference": JSON_SCHEMA_AAS_MODEL_REFERENCE,
            "ExternalReference": {
                "type": "string"
            }
        },
        "oneOf": [
            {"required": ["ModelReference"]},
            {"required": ["ExternalReference"]}
        ]
    }

    JSON_SCHEMA_ASSET_SERVICE_REQUEST = {
        "type": "object",
        "properties": {
            "serviceParameterValues": {
                "type": "object",
                "additionalProperties": {"type": "string"}
            },
            "ModelReference": JSON_SCHEMA_AAS_MODEL_REFERENCE,
        },
        "required": ["ModelReference"]
    }

    JSON_SCHEMA_CAPABILITY_REQUEST = {
        "type": "object",
        "properties": {
            "capabilityName": {"type": "string"},
            "skillName": {"type": "string"},
            "skillParameterValues": {
              "type": "object",
              "additionalProperties": {
                "type": "string"
              }
            },
            "skillInterfaceName": {"type": "string"}
        },
        "required": ["capabilityName"]
    }



"""
This is the SMIA launch file for Docker, which runs the program logic inside a virtualization container.
"""

import logging
import os

import smia
from smia.agents.smia_agent import SMIAAgent
from smia.utilities.general_utils import DockerUtils

_logger = logging.getLogger(__name__)

def main():
    # First, the initial configuration must be executed
    smia.initial_self_configuration()
    _logger.info("Initializing SMIA software...")

    # The AAS model is obtained from the environmental variables
    aas_model_path = DockerUtils.get_aas_model_from_env_var()

    # When the AAS model path has been obtained, it is added to SMIA
    smia.load_aas_model(aas_model_path)

    # The jid and password can also be set as environmental variables. In case they are not set, the values are obtained
    # from the initialization properties file
    smia_jid = os.environ.get('AGENT_ID')
    smia_psswd = os.environ.get('AGENT_PASSWD')

    # Create the agent object
    smia_agent = SMIAAgent(smia_jid, smia_psswd)
    smia.run(smia_agent)

if __name__ == '__main__':

    # Run main program with SMIA
    main()



import json
import logging

from spade.behaviour import CyclicBehaviour

from smia.behaviours.specific_handle_behaviours.handle_capability_behaviour import HandleCapabilityBehaviour
from smia.behaviours.specific_handle_behaviours.handle_svc_request_behaviour import HandleSvcRequestBehaviour
from smia.behaviours.specific_handle_behaviours.handle_svc_response_behaviour import HandleSvcResponseBehaviour
from smia.logic import inter_aas_interactions_utils
from smia.utilities.fipa_acl_info import FIPAACLInfo
from smia.utilities.general_utils import GeneralUtils

_logger = logging.getLogger(__name__)


class ACLHandlingBehaviour(CyclicBehaviour):
    """
    This class implements the behaviour that handles all the ACL messages that the SMIA will receive from the
    others standardized SMIAs in the I4.0 System.
    """

    def __init__(self, agent_object):
        """
        The constructor method is rewritten to add the object of the agent
        Args:
            agent_object (spade.Agent): the SPADE agent object of the SMIA agent.
        """

        # The constructor of the inherited class is executed.
        super().__init__()

        # The SPADE agent object is stored as a variable of the behaviour class
        self.myagent = agent_object

    async def on_start(self):
        """
        This method implements the initialization process of this behaviour.
        """
        _logger.info("ACLHandlingBehaviour starting...")

    async def run(self):
        """
        This method implements the logic of the behaviour.
        """

        # Wait for a message with the standard ACL template to arrive.
        msg = await self.receive(
            timeout=10)  # Timeout set to 10 seconds so as not to continuously execute the behavior.
        if msg:
            # An ACL message has been received by the agent
            _logger.aclinfo("         + Message received on SMIA (ACLHandlingBehaviour) from {}".format(msg.sender))
            _logger.aclinfo("                 |___ Message received with content: {}".format(msg.body))

            # The msg body will be parsed to a JSON object
            msg_json_body = json.loads(msg.body)

            # Depending on the performative of the message, the agent will have to perform some actions or others
            match msg.get_metadata('performative'):
                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_REQUEST:
                    _logger.aclinfo("The performative of the message is Request, so the DT needs to perform some"
                                    " action.")

                    # TODO pensar si generar un behaviour para recibir peticiones de servicios y otro para peticiones
                    #  de capacidades (en ese caso solo hay que comprobar la ontologia, no la performativa, ya que esta
                    #  se analiza en los behaviours de gestion individuales)
                    if ((msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_REQUEST) or
                            (msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING)):
                        _logger.aclinfo("The agent has received a request related to Capability-Skill model")
                        # The behaviour to handle this specific capability will be added to the agent
                        svc_req_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                        capability_handling_behav = HandleCapabilityBehaviour(self.agent, svc_req_data)
                        self.myagent.add_behaviour(capability_handling_behav)

                    elif msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_SVC_REQUEST:
                        _logger.aclinfo("The agent has received a request to perform a service")
                        # The behaviour to handle this specific capability will be added to the agent
                        svc_req_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                        svc_req_handling_behav = HandleSvcRequestBehaviour(self.agent, svc_req_data)
                        self.myagent.add_behaviour(svc_req_handling_behav)
                    else:
                        # TODO Parte del enfoque antiguo
                        service_category = msg_json_body['serviceData']['serviceCategory']
                        if service_category == 'service-response':
                            # TODO PENSAR COMO RECOGER LAS RESPUESTAS DE PETICIONES ANTERIORES. NO VENDRIAN CON
                            #  PERFORMATIVE 'Request', sai que aqui no se deberian tratar (recoger con Performative
                            #  'Inform', 'Confirm', 'Agree', 'Refuse'... dentro de los behaviours para gestiones
                            #  individuales?)
                            _logger.aclinfo(
                                "The agent has received the response of a service with thread [" + msg.thread + "].")

                            # As it is a response to a previous request, a new HandleSvcResponseBehaviour to handle this service
                            # response will be added to the agent
                            svc_resp_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                            svc_resp_handling_behav = HandleSvcResponseBehaviour(self.agent,
                                                                                 'Inter AAS interaction',
                                                                                 svc_resp_data)
                            self.myagent.add_behaviour(svc_resp_handling_behav)

                case FIPAACLInfo.FIPA_ACL_PERFORMATIVE_QUERY_IF:
                    _logger.aclinfo("The performative of the message is Query-If, so the DT has been asked about some "
                                    "aspect of it.")

                    # service_category = msg_json_body['serviceID']
                    # if service_category == 'capabilityChecking':
                    if msg.get_metadata('ontology') == FIPAACLInfo.FIPA_ACL_ONTOLOGY_CAPABILITY_CHECKING:
                        _logger.info("The DT has been asked to check if it has a given capability.")
                        svc_req_data = inter_aas_interactions_utils.create_svc_json_data_from_acl_msg(msg)
                        capability_handling_behav = HandleCapabilityBehaviour(self.agent, svc_req_data)
                        self.myagent.add_behaviour(capability_handling_behav)
                case _:
                    _logger.error("ACL performative type not available.")

        else:
            _logger.info("         - No message received within 10 seconds on SMIA (ACLHandlingBehaviour)")



"""
This class contains methods related to negotiation management of the SMIA.
"""
import json
import logging

from spade.message import Message

from smia.utilities.smia_info import SMIAInteractionInfo
from smia.utilities.general_utils import GeneralUtils

_logger = logging.getLogger(__name__)


def create_neg_cfp_msg(thread, targets, neg_requester_jid, neg_criteria):
    """
    This method creates the FIPA-ACL CallForProposal (CFP) message that will be sent to all participants to start
    a negotiation.

    Args:
        thread (str): the thread of the ACL message.
        targets (str): the JIDs of the SPADE agents that are participating in the negotiation. Is it a string that has the JIDs divided by ','
        neg_requester_jid (str):  the JID of the SPADE agent that has requested the negotiation.
        neg_criteria (str): criteria of the negotiation.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-compliant.
    """
    cfp_msg = Message(thread=thread)
    cfp_msg.metadata = SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE_CFP.metadata

    neg_cfp_json = {
        'serviceID': 'startNegotiation',
        'serviceType': 'AssetRelatedService',  # TODO cambiarlo si se decide que es de otro tipo
        'serviceData': {
            'serviceCategory': 'service-request',
            'timestamp': GeneralUtils.get_current_timestamp(),
            'serviceParams': {
                'targets': targets,
                'neg_requester_jid': neg_requester_jid,
                'criteria': neg_criteria
            }
        }
    }
    cfp_msg.body = json.dumps(neg_cfp_json)
    return cfp_msg


def create_neg_propose_msg(thread, targets, neg_requester_jid, neg_criteria, neg_value):
    """
    This method creates the FIPA-ACL propose message that will be sent to all participants in a negotiation to check
    their negotiation values against the value of this SMIA.

    Args:
        thread (str): the thread of the ACL message.
        targets (str): the JIDs of the SPADE agents that are participating in the negotiation. Is it a string that has the JIDs divided by ','
        neg_requester_jid (str):  the JID of the SPADE agent that has requested the negotiation.
        neg_criteria (str): criteria of the negotiation.
        neg_value (str): value of the negotiation.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-compliant.
    """
    propose_msg = Message(thread=thread)
    propose_msg.metadata = SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE_PROPOSE.metadata

    neg_propose_json = {
        'serviceID': 'proposeNegotiation',
        'serviceType': 'AssetRelatedService',  # TODO cambiarlo si se decide que es de otro tipo
        'serviceData': {
            'serviceCategory': 'service-request',
            'timestamp': GeneralUtils.get_current_timestamp(),
            'serviceParams': {
                'targets': targets,
                'neg_requester_jid': neg_requester_jid,
                'criteria': neg_criteria,
                'neg_value': neg_value
            }
        }
    }
    propose_msg.body = json.dumps(neg_propose_json)
    return propose_msg


def create_neg_response_msg(receiver, thread, service_id, service_type, winner):
    """
    This method creates the FIPA-ACL message to respond to a negotiation request with its result.

    Args:
        receiver (str): the JID of the receiver of the ACL message, that it will be the requester of the negotiation.
        thread (str): the thread of the ACL message.
        service_id (str): the serviceID of the ACL message.
        service_type (str): the serviceType of the ACL message.
        winner (str): the JID of the winner of the negotiation.

    Returns:
        spade.message.Message: SPADE message object FIPA-ACL-compliant.
    """
    response_msg = Message(to=receiver, thread=thread)
    response_msg.metadata = SMIAInteractionInfo.NEG_STANDARD_ACL_TEMPLATE_INFORM.metadata

    neg_response_json = {
        'serviceID': service_id,
        'serviceType': service_type,
        'serviceData': {
            'serviceCategory': 'service-response',
            'serviceStatus': 'Completed',
            'timestamp': GeneralUtils.get_current_timestamp(),
            'serviceParams': {
                'winner': winner
            }
        }
    }
    response_msg.body = json.dumps(neg_response_json)
    return response_msg


def create_neg_json_to_store(neg_requester_jid, participants, neg_criteria, is_winner):
    """
    This method creates the JSON object to be stored in the global object of the SMIA for the information of all
    negotiations in which it has participated.

    Args:
        neg_requester_jid (str):  the JID of the SPADE agent that has requested the negotiation.
        participants (str): JIDs of the SPADE agents that have participated in the negotiation. Is it a string that has the JIDs divided by ','
        neg_criteria (str): criteria of the negotiation.
        is_winner (bool): it determines whether the SMIA has been the winner of the negotiation.

    Returns:
        dict: object with all the information of the negotiation in JSON format.
    """
    return {
        'neg_requester_jid': neg_requester_jid,
        'participants': participants,
        'neg_criteria': neg_criteria,
        'is_winner': str(is_winner)
    }


def create_intra_aas_neg_req_data(performative, ontology, thread, service_data):
    """
    This method creates the dictionary with all the required data for an Intra AAS interaction request.

    Args:
        performative (str): performative of the FIPA-ACL message.
        ontology (str): ontology of the FIPA-ACL message.
        thread (str): thread of the FIPA-ACL message.
        service_data (dict): all the data of the service for the Intra AAS interaction request.

    Returns:
        dict: dictionary with all the information about the Intra AAS interaction request
    """
    intra_ass_req_data_json = {
        'performative': performative,
        'ontology': ontology,
        'thread': thread,
        'serviceID': 'getNegotiationValue',
        'serviceType': 'AssetRelatedService',
        'serviceData': service_data
    }
    return intra_ass_req_data_json


def add_value_and_unlock_neg_handling_behaviour(agent, thread, neg_value):
    """
    This method adds the value to the HandleNegotiationBehaviour which is in charge of the negotiation with the given
    thread. Since this behaviour is waiting for this value, this method also unlocks its execution, so it can continue
    with its logic.

    Args:
        agent (spade.agent.Agent): the SPADE agent object that represents the SMIA.
        thread (str): the thread of the negotiation, which is its identifier.
        neg_value (int): value of the AAS in the negotiation-
    """
    # First, it will get the exact behaviour responsible for the negotiation for the given thread
    for behaviour in agent.behaviours:
        behav_class_name = str(behaviour.__class__.__name__)
        if behav_class_name == 'HandleNegotiationBehaviour':
            if behaviour.thread == thread:
                # Once the exact behaviour has been found, the negotiation value will be stored as an attribute, and the
                # method will be concluded
                behaviour.neg_value = float(neg_value)
                # The execution of this behaviour is also unlocked
                behaviour.neg_value_event.set()
                break


def get_neg_intra_aas_request_by_thread(agent, thread):
    """
    This method gets the data of an Intra AAS Interaction using the thread value.
    Args:
        agent (agents.smia_agent): the SPADE agent object that represents the SMIA.
        thread (str): the thread of the negotiation, which is its identifier.

    Returns:
        dict: dictionary with all the information about the Intra AAS interaction request. None if it does not exist.
    """
    for req_interaction_id, req_data in agent.interaction_requests.items():
        if req_data['thread'] == thread:
            req_data['interactionID'] = req_interaction_id
            return req_data
    return None



from basyx.aas.model import AssetAdministrationShell, AssetInformation, Submodel, SubmodelElement, RelationshipElement, \
    AnnotatedRelationshipElement, Capability, Operation, BasicEventElement, Entity, SubmodelElementList, \
    SubmodelElementCollection, Property, MultiLanguageProperty, Range, Blob, File, ReferenceElement, \
    ConceptDescription, Qualifier, HasSemantics, Namespace, Qualifiable


class AASModelExtensionUtils:
    """
    This class contains utility methods related to the extension of the BaSyx Python SDK AAS model.
    """

    @staticmethod
    def get_extension_classes_dict():
        """
        This method returns the dictionary with the link between all BaSyx AAS classes and Extended SMIA classes.

        Returns:
            dict: dictionary with the BaSyx classes as keys and Extended SMIA classes as values.
        """
        # The libraries are imported locally to avoid circular import error
        from smia.aas_model.extended_aas import ExtendedAssetAdministrationShell, ExtendedAssetInformation
        from smia.aas_model.extended_base import ExtendedQualifier, ExtendedHasSemantics, ExtendedNamespace, \
            ExtendedQualifiable
        from smia.aas_model.extended_concept_description import ExtendedConceptDescription
        from smia.aas_model.extended_submodel import ExtendedSubmodel, ExtendedSubmodelElement, ExtendedRelationshipElement, \
            ExtendedAnnotatedRelationshipElement, ExtendedCapability, ExtendedOperation, ExtendedBasicEventElement, \
            ExtendedEntity, ExtendedSubmodelElementList, ExtendedSubmodelElementCollection, ExtendedProperty, \
            ExtendedMultiLanguageProperty, ExtendedRange, ExtendedBlob, ExtendedFile, ExtendedReferenceElement

        return {AssetAdministrationShell: ExtendedAssetAdministrationShell,
                AssetInformation: ExtendedAssetInformation,
                Submodel: ExtendedSubmodel,
                SubmodelElement: ExtendedSubmodelElement,
                RelationshipElement: ExtendedRelationshipElement,
                AnnotatedRelationshipElement: ExtendedAnnotatedRelationshipElement,
                Capability: ExtendedCapability,
                Operation: ExtendedOperation,
                BasicEventElement: ExtendedBasicEventElement,
                Entity: ExtendedEntity,
                SubmodelElementList: ExtendedSubmodelElementList,
                SubmodelElementCollection: ExtendedSubmodelElementCollection,
                Property: ExtendedProperty,
                MultiLanguageProperty: ExtendedMultiLanguageProperty,
                Range: ExtendedRange,
                Blob: ExtendedBlob,
                File: ExtendedFile,
                ReferenceElement: ExtendedReferenceElement,
                ConceptDescription: ExtendedConceptDescription,
                Qualifier: ExtendedQualifier,
                HasSemantics: ExtendedHasSemantics,
                Namespace: ExtendedNamespace,
                Qualifiable: ExtendedQualifiable
                }


    @staticmethod
    def extend_basyx_aas_model():
        for model_class, extension_class in AASModelExtensionUtils.get_extension_classes_dict().items():
            for method_name in dir(extension_class):
                if callable(getattr(extension_class, method_name)) and not method_name.startswith("__"):
                    # Special methods (e.g., __init__) are excluded.
                    setattr(model_class, method_name, getattr(extension_class, method_name))




"""
This class contains methods related to service management. It contains all type of services proposed in the
Functional View of RAMI 4.0.
"""
import inspect
import logging
import types

_logger = logging.getLogger(__name__)

class AgentServiceUtils:
    """
    This class contains utility methods related to the Agent Services.
    """

    @staticmethod
    async def get_agent_service_parameters(service_method):
        """
        This method gets the required information about the parameters of the agent service: the names and types of
        service method parameters.

        Args:
            service_method (method): method of the agent service.

        Returns:
            dict: dictionary with all information about the parameters of the agent service.
        """
        if service_method is None:
            raise KeyError(f"The service method {service_method} does not exist in this DT.")
        else:
            # Obtain the signature of the method
            sig = inspect.signature(service_method)
            params = sig.parameters
            # Construct the dictionary with the information of the parameters
            params_details = {
                nombre: {
                    'annotation': parametro.annotation,
                    'type': parametro.annotation if parametro.annotation is not inspect.Parameter.empty else str,
                }
                for nombre, parametro in params.items()
            }
            return params_details


    @staticmethod
    async def get_adapted_service_parameters(service_method, **kwargs):
        """
        This method adapts the received parameters values to the required types of the service method, in order to be
        correctly executed.

        Args:
            service_method (method): method of the agent service.
            **kwargs: received parameters with the values.

        Returns:
            dict: parameters correctly adapted to the method.
        """
        if service_method is None:
            ValueError(f"A null object has been offered for the {service_method} method, its parameters cannot be "
                       f"obtained.")
        # The information of the required method parameters is obtained
        required_params_info = await AgentServiceUtils.get_agent_service_parameters(service_method)
        adapted_params = {}

        # The received parameters with the values are available in kwargs
        if len(required_params_info) != 0 and ((kwargs is None) or (len(kwargs) == 0)):
            raise ValueError(f"The service method cannot be executed because the required parameters have not been "
                             f"provided.")

        for param_name, value in kwargs.items():
            if param_name in required_params_info:
                tipo = required_params_info[param_name]['type']
                try:
                    if tipo == bool:  # bool(value) is true as long as the string is not empty, son it cannot be used.
                        adapted_params[param_name] = value.lower() in ('yes', 'true', 't', '1')
                    else:
                        adapted_params[param_name] = tipo(value)
                except ValueError as e:
                    raise ValueError(f"Could not transform the value {value} for the parameter '{param_name}',"
                                     f" reason: {e}")
            else:
                raise ValueError(f"Parameter {param_name} not found in method {service_method}.")
        return adapted_params

    @staticmethod
    async def safe_execute_agent_service(service_method, **kwargs):
        """
        This method executes the agent service securely, regardless of the execution method, synchronous or
        asynchronous.

        Args:
            service_method: executable method of the agent service.
            **kwargs: in case the execution method has parameters, they are passed as kwargs.

        Returns:
            result of the execution of the agent service.
        """
        if inspect.iscoroutinefunction(service_method):
            return await service_method(**kwargs)
        else:
            return service_method(**kwargs)


class SubmodelServicesUtils:
    pass



import logging
from spade.behaviour import FSMBehaviour

_logger = logging.getLogger(__name__)


class AASFSMBehaviour(FSMBehaviour):
    """
    This class is an extension of the own FMSBehaviour of SPADE, which allows to print the start and the finish of
    the Finite State Machine.
    """
    async def on_start(self):
        """This method prints that the FSM has started."""
        _logger.info(f"    ** FSM starting at initial state --> {self.current_state}")

    async def on_end(self):
        """This method prints that the FSM has stopped and execute the agent-related stopping method."""
        _logger.info(f"    ** FSM finished at state --> {self.current_state}")
        await self.agent.stop()
